def set_value(config, key, value):
    key_list = key.split(".")
    handle = config
    for keypart in key_list[:-1]:
        handle = handle[keypart]
    handle[key_list[-1]] = value


def index_key(tree, key):
    if tree is None:
        return False, None

    key_list = key.split(".")
    handle = tree
    for keypart in key_list:
        if keypart not in handle:
            return False, None
        handle = handle[keypart]
    return True, handle


def del_value(config, key):
    key_list = key.split(".")
    handle = config
    traceback = []
    for keypart in key_list[:-1]:
        traceback.append(handle)
        handle = handle[keypart]
    if key_list[-1] in handle:
        del handle[key_list[-1]]
    
    # cleanup empty
    if len(key_list) > 1:
        for off, keypart in zip(range(len(key_list) - 2, -1, -1), key_list[-2::-1]):
            handle_curr = traceback[off]
            if len(handle_curr[keypart]) == 0:
                del handle_curr[keypart]


from __future__ import annotations

import argparse
import enum
from enum import auto
import re

import typing
from typing import Any

if typing.TYPE_CHECKING:
    from .reg import ConfigEntryAttr


@enum.unique
class ConfigEntrySource(enum.IntEnum):
    BUILTIN = 0
    CALLBACK = auto()
    COMMANDLINE_ONLY = auto()
    CONFIG_ONLY = auto()
    COMMANDLINE_OVER_CONFIG = auto()


class ConfigEntryCommandlinePattern:
    pass


@enum.unique
class ConfigEntryCommandlineSeqPattern(ConfigEntryCommandlinePattern, enum.IntEnum):
    COMMA_SEP = 0
    COLON_SEP = auto()
    SEMICOLON_SEP = auto()


@enum.unique
class ConfigEntryCommandlineMapPattern(ConfigEntryCommandlinePattern, enum.IntEnum):
    COMMA_SEP_EQUAL = 0
    COLON_SEP_EQUAL = auto()
    SEMICOLON_SEP_EQUAL = auto()
    COMMA_SEP_COLON = auto()
    SEMICOLON_SEP_COLON = auto()


@enum.unique
class ConfigEntryCommandlineBoolPattern(ConfigEntryCommandlinePattern, enum.IntEnum):
    SET_TRUE = 0
    SET_FALSE = auto()
    ON_OFF = auto()


class ConfigEntryValueUnspecified:
    pass


def analyze_type(in_type, seq_type_list, map_type_list):

    def _analyze_type(_in_type, _seq_type_list, _map_type_list):
        if _in_type == Any:
            return {"proc": None, "cast": None, "next": []}

        origin = typing.get_origin(_in_type)
        if origin is None:
            return {"proc": _in_type, "cast": None, "next": []}

        if origin in _seq_type_list:
            args = typing.get_args(_in_type)
            if len(args) != 1:
                raise TypeError(f"type parameter error! got {args} with len {len(args)}")
            argtype = args[0]
            extra_proc = _analyze_type(argtype, _seq_type_list, _map_type_list)
            return {"proc": origin, "cast": "seq", "next": [extra_proc]}

        if origin in _map_type_list:
            args = typing.get_args(_in_type)
            if len(args) not in [1, 2]:
                raise TypeError(f"type parameter error! got {args} with len {len(args)}")
            if len(args) == 1 and args[0] == Any:
                args = (Any, Any)
            keyargtype = args[0]
            key_extra_proc = _analyze_type(keyargtype, _seq_type_list, _map_type_list)
            next_list = [key_extra_proc]
            if len(args) == 2:
                valueargtype = args[1]
                value_extra_proc = _analyze_type(valueargtype, _seq_type_list, _map_type_list)
                next_list.append(value_extra_proc)
            else:
                next_list.append({"proc": None, "cast": None, "next": []})
            return {"proc": origin, "cast": "map", "next": next_list}

        raise TypeError(f"unsupported container type! got {origin}")

    return _analyze_type(in_type, seq_type_list, map_type_list)


def supported_by_commandline(proclist):
    if proclist["cast"] == "seq":
        return len(proclist["next"][0]["next"]) == 0
    if proclist["cast"] == "map":
        return (len(proclist["next"][0]["next"]) == 0) and (len(proclist["next"][1]["next"]) == 0)
    return True


def proclist_pattern_paired(proclist, cmdpattern, supported_seq, supported_map):
    if proclist["cast"] == "seq":
        return isinstance(cmdpattern, ConfigEntryCommandlineSeqPattern)
    if proclist["cast"] == "map":
        return isinstance(cmdpattern, ConfigEntryCommandlineMapPattern)
    if proclist["proc"] in supported_seq:
        return isinstance(cmdpattern, ConfigEntryCommandlineSeqPattern)
    if proclist["proc"] in supported_map:
        return isinstance(cmdpattern, ConfigEntryCommandlineMapPattern)
    if proclist["proc"] == bool:
        return isinstance(cmdpattern, ConfigEntryCommandlineBoolPattern)
    return cmdpattern is None


def cast_to_res(blob, proclist):

    def _cast_to_res(_blob, _proclist):
        _res = _blob
        if _proclist["cast"] == "seq":
            # handle str, int (single element case)
            if isinstance(_blob, str):
                # raise TypeError(f"config contain str in list field!")
                _res = [_blob]
            elif isinstance(_blob, int):
                _res = [_blob]
            # TODO: support more elementary type?
            else:
                _res = list(_blob)
            next_list = _proclist["next"]
            if len(next_list) != 1:
                raise ValueError(f"Expected exactly 1 next type for sequence, got {len(next_list)}")
            next_proclist = next_list[0]
            for offset in range(len(_res)):
                _res[offset] = _cast_to_res(_res[offset], next_proclist)
        elif _proclist["cast"] == "map":
            _res = dict(_blob)
            next_list = _proclist["next"]
            if len(next_list) != 2:
                raise ValueError(f"Expected exactly 2 next types for map (key, value), got {len(next_list)}")
            _res2 = _res
            _res = {}
            for k, v in _res2.items():
                new_k = _cast_to_res(k, next_list[0])
                new_v = _cast_to_res(v, next_list[1])
                _res[new_k] = new_v

        if _proclist["proc"] is not None:
            _res = _proclist["proc"](_res)

        return _res

    return _cast_to_res(blob, proclist)


match_special = re.compile(r"\?\((.*?)\)")


def split_ignore_special(in_str, sep):
    # split the in_str with sep, but ignore sep that inside match of match_special
    # e.g. sep=:
    # in_str=a:b -> ["a", "b"]
    # in_str=?(a:c):b -> ["?(a:c)", "b"]
    # in_str=?(a:c):?(b:d) -> ["?(a:c)", "?(b:d)"]
    # in_str=zz?(a:c):b -> [zz?(a:c)", "b"]

    res = []
    base, cur = 0, 0
    while True:
        if base >= len(in_str):
            break

        if cur >= len(in_str) or in_str[cur] == sep:  # find a separator
            res.append(in_str[base:cur])
            base = cur + 1
            cur = base
        else:
            match = match_special.search(in_str, cur)
            if match:
                start, end = match.span()
                cur = end
            else:
                cur += 1
    return res


def handle_cmd_seq(in_str, sep_type):
    if sep_type == ConfigEntryCommandlineSeqPattern.COMMA_SEP:
        sep = ","
    elif sep_type == ConfigEntryCommandlineSeqPattern.COLON_SEP:
        sep = ":"
    elif sep_type == ConfigEntryCommandlineSeqPattern.SEMICOLON_SEP:
        sep = ";"
    else:
        raise TypeError(f"unknown cmdline seq pattern! got {sep_type}")

    res = split_ignore_special(in_str, sep)
    return res


def handle_cmd_map(in_str, sep_type):
    if sep_type == ConfigEntryCommandlineMapPattern.COMMA_SEP_EQUAL:
        sep_pair, sep_eq = ",", "="
    elif sep_type == ConfigEntryCommandlineMapPattern.COLON_SEP_EQUAL:
        sep_pair, sep_eq = ":", "="
    elif sep_type == ConfigEntryCommandlineMapPattern.SEMICOLON_SEP_EQUAL:
        sep_pair, sep_eq = ";", "="
    elif sep_type == ConfigEntryCommandlineMapPattern.COMMA_SEP_COLON:
        sep_pair, sep_eq = ",", ":"
    elif sep_type == ConfigEntryCommandlineMapPattern.SEMICOLON_SEP_COLON:
        sep_pair, sep_eq = ";", ":"
    else:
        raise TypeError(f"unknown cmdline map pattern! got {sep_type}")

    base_list = in_str.split(sep_pair)
    res = {}
    for base_pair in base_list:
        k, v = base_pair.split(sep_eq)
        res[k] = v
    return res


def hook_cmd_bool(parser: argparse.ArgumentParser, key: str, pattern: ConfigEntryCommandlineBoolPattern,
                  meta: ConfigEntryAttr):
    if pattern == ConfigEntryCommandlineBoolPattern.SET_TRUE:
        parser.add_argument(f"--{key}", action="store_true", default=argparse.SUPPRESS, help=meta.desc)
    elif pattern == ConfigEntryCommandlineBoolPattern.SET_FALSE:
        parser.add_argument(f"--{key}", action="store_false", default=argparse.SUPPRESS, help=meta.desc)
    elif pattern == ConfigEntryCommandlineBoolPattern.ON_OFF:
        parser.add_argument(f"--{key}", action="store_true", default=argparse.SUPPRESS, help=meta.desc)
        opt_str = list(parser._option_string_actions.keys())
        if f"--{key}__off" in opt_str:
            raise KeyError(f"parser already have string action `--{key}__off`!")
        off_desc = meta.desc + " (off)" if meta.desc is not None else None
        parser.add_argument(f"--{key}__off", action="store_true", default=argparse.SUPPRESS, help=off_desc)
    else:
        raise TypeError(f"unknown cmdline bool pattern! got {pattern}")


def handle_cmd_bool(parse_res: argparse.Namespace, key: str, pattern: ConfigEntryCommandlineBoolPattern):
    if pattern == ConfigEntryCommandlineBoolPattern.SET_TRUE:
        # Use hasattr to check if arg exists (due to SUPPRESS)
        if hasattr(parse_res, key) and getattr(parse_res, key):
            res = True
        else:
            res = ConfigEntryValueUnspecified
    elif pattern == ConfigEntryCommandlineBoolPattern.SET_FALSE:
        # For SET_FALSE, only set when arg exists and is False
        if hasattr(parse_res, key) and not getattr(parse_res, key):
            res = False
        else:
            res = ConfigEntryValueUnspecified
    elif pattern == ConfigEntryCommandlineBoolPattern.ON_OFF:
        on_flag = getattr(parse_res, key, False)
        off_flag = getattr(parse_res, f"{key}__off", False)
        if not on_flag and not off_flag:
            res = ConfigEntryValueUnspecified
        elif on_flag and not off_flag:
            res = True
        elif not on_flag and off_flag:
            res = False
        else:
            raise RuntimeError(f"both on and off flag are set! key: {key}")
    else:
        raise TypeError(f"unknown cmdline bool pattern! got {pattern}")

    return res



import re

from typing import Optional


match_special = re.compile(r"^\?\((.*)\)$")


def extract_special(s: str) -> Optional[str]:
    """Extract content from special pattern ?(content).
    
    Args:
        s: String to check for special pattern.
        
    Returns:
        The content inside ?() if matched, None otherwise.
    """
    match_special_res = match_special.fullmatch(s)
    if match_special_res:
        return match_special_res.group(1)
    else:
        return None


match_special_part = re.compile(r"\?\((.*?)\)")


def extract_special_part(s: str) -> tuple[list[str], list[tuple[int, int]]]:
    """Extract all special patterns ?(content) from a string.
    
    Args:
        s: String to search for special patterns.
        
    Returns:
        A tuple of (cmd_list, span_list) where cmd_list contains the extracted
        content and span_list contains the (start, end) positions.
    """
    cmd_list: list[str] = []
    span_list: list[tuple[int, int]] = []
    for match in match_special_part.finditer(s):
        cmd = match.group(1)
        span = match.span()
        cmd_list.append(cmd)
        span_list.append(span)
    return cmd_list, span_list


def replace_from_span(s: str, span_list: list[tuple[int, int]], replacement_list: list[str]) -> str:
    """Replace spans in a string with replacements.
    
    Args:
        s: The original string.
        span_list: List of (start, end) index tuples for each span in the original string.
        replacement_list: List of replacement strings, one for each span.
        
    Returns:
        A new string with all spans replaced.
        
    Raises:
        ValueError: If span_list and replacement_list have different lengths.
    """
    if len(span_list) != len(replacement_list):
        raise ValueError(f"span_list and replacement_list must have same length, got {len(span_list)} and {len(replacement_list)}")
    # sort by start index
    span_list = sorted(span_list, key=lambda x: x[0])
    # replace from end to start
    for i in range(len(span_list) - 1, -1, -1):
        start, end = span_list[i]
        s = s[:start] + replacement_list[i] + s[end:]
    return s


match_file = re.compile(r"^file:(.*)$")


def extract_file(s: str) -> Optional[str]:
    """Extract filename from file: prefix pattern.
    
    Args:
        s: String to check for file: pattern.
        
    Returns:
        The filename if matched, None otherwise.
    """
    match_file_res = match_file.fullmatch(s)
    if match_file_res:
        return match_file_res.group(1)
    else:
        return None



from __future__ import annotations

import logging
import os
import sys
import argparse
from typing import Optional, Union, Any
from typing import Sequence, Mapping
from dataclasses import dataclass

from .type_def import ConfigEntrySource, ConfigEntryValueUnspecified, analyze_type, cast_to_res
from .type_def import (
    ConfigEntryCommandlinePattern,
    ConfigEntryCommandlineSeqPattern,
    ConfigEntryCommandlineMapPattern,
)
from .type_def import supported_by_commandline, proclist_pattern_paired
from .type_def import handle_cmd_seq, handle_cmd_map
from .type_def import hook_cmd_bool, handle_cmd_bool
from .type_transform import is_type

from .index import set_value, index_key, del_value

from .callback import ConfigEntryCallback, resolve_callback_dependency

from .if_yaml import load_yaml

from copy import deepcopy

logger = logging.getLogger(__name__)


def split_argv_by_cfg(argv: list[str]) -> list[tuple[str, any]]:
    """
    Split argv by -c/--cfg markers into segments.

    The '--' marker stops config file parsing; all arguments after '--' are
    treated as regular arguments and '-c/--cfg' will not be interpreted as
    config file markers.

    Returns: [(segment_type, content), ...]
        segment_type: 'args' | 'cfg'
        content:
            - 'args': list[str] - argument list
            - 'cfg': str - config file path

    Example:
        Input: ['--a', '1', '-c', 'f1.yaml', '--b', '2', '--cfg=f2.yaml', '--c', '3']
        Output: [
            ('args', ['--a', '1']),
            ('cfg', 'f1.yaml'),
            ('args', ['--b', '2']),
            ('cfg', 'f2.yaml'),
            ('args', ['--c', '3'])
        ]

        Input: ['--a', '1', '--', '-c', 'not_a_config']
        Output: [
            ('args', ['--a', '1', '--', '-c', 'not_a_config'])
        ]
    """
    segments = []
    current_args = []
    i = 0
    stop_cfg_parsing = False  # Set to True after encountering '--'

    while i < len(argv):
        arg = argv[i]

        # Check for '--' terminator
        if arg == '--':
            current_args.append(arg)
            stop_cfg_parsing = True
            i += 1
            continue

        # After '--', treat everything as regular arguments
        if stop_cfg_parsing:
            # Warn if -c/--cfg appears after '--'
            if arg == '-c' or arg == '--cfg' or arg.startswith('--cfg='):
                logger.warning(f"'{arg}' after '--' will be treated as regular argument, "
                               "not as config file marker")
            current_args.append(arg)
            i += 1
            continue

        # Case 1: -c <file> or --cfg <file> (two tokens)
        if arg == '-c' or arg == '--cfg':
            # Save current accumulated args segment
            if current_args:
                segments.append(('args', current_args))
                current_args = []

            # Get config file path
            if i + 1 < len(argv):
                cfg_file = argv[i + 1]

                # Warn if config file path looks like an option (starts with -)
                # This may indicate the previous argument expected a value
                if cfg_file.startswith('-'):
                    logger.warning(f"Config file path '{cfg_file}' looks like an option. "
                                   f"Did you mean to pass '{cfg_file}' as a value to the previous argument? "
                                   "Use '--' to stop config file parsing if needed.")

                segments.append(('cfg', cfg_file))
                i += 2  # Skip -c and file path
            else:
                raise ValueError(f"Missing config file path after {arg}")

        # Case 2: --cfg=<file> (single token)
        elif arg.startswith('--cfg='):
            if current_args:
                segments.append(('args', current_args))
                current_args = []

            cfg_file = arg[6:]  # Remove '--cfg='
            if not cfg_file:
                raise ValueError("Empty config file path in --cfg=")

            # Warn if config file path looks like an option
            if cfg_file.startswith('-'):
                logger.warning(f"Config file path '{cfg_file}' looks like an option. "
                               "Use '--' to stop config file parsing if needed.")

            segments.append(('cfg', cfg_file))
            i += 1

        # Case 3: regular argument
        else:
            current_args.append(arg)
            i += 1

    # Add final args segment
    if current_args:
        segments.append(('args', current_args))

    return segments


@dataclass
class ConfigEntryAttr:
    source: ConfigEntrySource
    desc: Optional[str]
    category: Union[type, Any]
    cmdpattern: Optional[ConfigEntryCommandlinePattern]
    default: Any
    callback: Optional[ConfigEntryCallback]
    required: bool


def prepare_default_config(meta_info_tree):

    def _prepare_default_config(_meta_info_tree):
        if isinstance(_meta_info_tree, ConfigEntryAttr):
            return deepcopy(_meta_info_tree.default)

        res = {}
        for key in _meta_info_tree:
            res[key] = _prepare_default_config(_meta_info_tree[key])
        return res

    return _prepare_default_config(meta_info_tree)


def check_config_integrity(meta_info_tree, config_tree, prefix=None):

    def _check_config_integrity(_meta_info_tree, _config_tree, _prefix=None):
        if isinstance(_meta_info_tree, ConfigEntryAttr):
            if _config_tree != ConfigEntryValueUnspecified:
                return True, None
            elif not _meta_info_tree.required:
                return True, [_prefix]
            else:
                return False, [_prefix]

        res = True
        res_list = []
        for key in _meta_info_tree:
            if _prefix is None:
                new_prefix = key
            else:
                new_prefix = _prefix + "." + key
            status, problem_list = _check_config_integrity(_meta_info_tree[key], _config_tree[key], new_prefix)
            if not status:
                res = False
                res_list.extend(problem_list)
        return res, res_list

    return _check_config_integrity(meta_info_tree, config_tree, prefix)


class ConfigRegistry:

    def __init__(self, prog: str = "prog", warn_unknown_key: bool = False):
        self.prog = prog
        self.warn_unknown_key = warn_unknown_key

        self.supported_seq_type: list[type] = [list, tuple]
        self.supported_map_type: list[type] = [dict]

        self.reset()

    def reset(self):
        # meta: store key info
        self.meta_info: dict[str, ConfigEntryAttr] = {}
        self.meta_info_tree: dict[str, Any] = {}
        self.category_proclist_cache: dict[Union[type, Any], Any] = {}

        # bind: store
        self.bind_config_filepath_list: list[str] = []

        # config: store parse res
        self.config = {}
        self.lack_key_list = []

    def reg_seq_type(self, newtype: type) -> None:
        if not isinstance(newtype, type):
            raise TypeError(f"Expected a type, got {type(newtype).__name__}: {newtype}")
        self.supported_seq_type.append(newtype)

    def reg_map_type(self, newtype: type) -> None:
        if not isinstance(newtype, type):
            raise TypeError(f"Expected a type, got {type(newtype).__name__}: {newtype}")
        self.supported_map_type.append(newtype)

    def register(
        self,
        key: str,
        prefix: Optional[str] = None,
        category: Union[type, Any] = Any,
        source: ConfigEntrySource = ConfigEntrySource.BUILTIN,
        desc: Optional[str] = None,
        required: bool = False,
        default: Any = ConfigEntryValueUnspecified,
        cmdpattern: Optional[ConfigEntryCommandlinePattern] = None,
        callback: Optional[ConfigEntryCallback] = None,
    ):
        # check key
        if key is None or len(key) == 0:
            return
        if key == "cfg":
            raise KeyError("collide with internal key!")

        # check cata
        if category != Any and not is_type(category):
            raise TypeError(f"error in category, got {category}")
        if category not in self.category_proclist_cache:
            proclist = analyze_type(category, self.supported_seq_type, self.supported_map_type)
            self.category_proclist_cache[category] = proclist

        # check src
        if not isinstance(source, ConfigEntrySource):
            raise TypeError(f"error in source, got {type(source)}")
        if source in [ConfigEntrySource.COMMANDLINE_ONLY, ConfigEntrySource.COMMANDLINE_OVER_CONFIG]:
            if not supported_by_commandline(self.category_proclist_cache[category]):
                raise TypeError(f"category not supported! got category {category}")
            if not proclist_pattern_paired(self.category_proclist_cache[category], cmdpattern, self.supported_seq_type,
                                           self.supported_map_type):
                raise TypeError(f"proclist & cmdpattern not paired! got category {category} and pattern {cmdpattern}")

        # key_list
        key_list = key.split(".")
        if prefix is not None:
            _key_list = prefix.split(".")
            key_list = _key_list + key_list
        internal_key = ".".join(key_list)

        # record
        attr_tuple = ConfigEntryAttr(
            source=source,
            desc=desc,
            category=category,
            cmdpattern=cmdpattern,
            default=default,
            callback=callback,
            required=required,
        )
        _handle = self.meta_info_tree
        for offset, key_part in enumerate(key_list):
            if key_part in _handle:
                if isinstance(_handle[key_part], ConfigEntryAttr):
                    raise KeyError(f"conflict prefix and key at {key_list[:offset+1]}")
                elif isinstance(_handle[key_part], dict) and offset == len(key_list) - 1:
                    raise KeyError(f"conflict prefix and key at {key_list[:offset+1]}")
                _handle = _handle[key_part]
            else:
                if offset == len(key_list) - 1:
                    _handle[key_part] = attr_tuple
                else:
                    _handle[key_part] = {}
                    _handle = _handle[key_part]
        self.meta_info[internal_key] = attr_tuple

        return self

    def bind_default_config_filepath(self, cfg_filepath: Union[str, Sequence[str]]):
        if isinstance(cfg_filepath, str):
            self.bind_config_filepath_list.append(cfg_filepath)
        else:
            self.bind_config_filepath_list.extend(cfg_filepath)

    def hook(self, parser: Optional[argparse.ArgumentParser] = None):
        if parser is None:
            return

        self.hook_config(parser)
        self.hook_arg(parser)

    def hook_config(self, parser: Optional[argparse.ArgumentParser] = None):
        # In segmented parsing, -c/--cfg is handled by split_argv_by_cfg() beforehand.
        # We still register a dummy option for:
        # 1. Displaying usage in --help
        # 2. Preventing users from registering conflicting -c/--cfg arguments
        if parser is None:
            return

        # Check if -c/--cfg already exists
        opt_str = list(parser._option_string_actions.keys())
        if "-c" in opt_str or "--cfg" in opt_str:
            raise KeyError("parser already have string action `-c` or `--cfg`!")

        # Register dummy option (actually handled by split_argv_by_cfg, never parsed by argparse)
        parser.add_argument(
            "-c",
            "--cfg",
            action="append",
            default=argparse.SUPPRESS,
            metavar="FILE",
            help=f"{self.__class__.__name__} config file (can be interleaved with options in any order)")

    def hook_arg(self, parser: Optional[argparse.ArgumentParser] = None):
        if parser is None:
            return

        # hook arg
        entry_cmdline = list(entry_key for entry_key, entry_meta in self.meta_info.items()
                             if entry_meta.source in (ConfigEntrySource.COMMANDLINE_ONLY,
                                                      ConfigEntrySource.COMMANDLINE_OVER_CONFIG))
        for entry_key in entry_cmdline:
            entry_meta = self.meta_info[entry_key]
            if entry_meta.category == bool:
                hook_cmd_bool(parser, entry_key, entry_meta.cmdpattern, entry_meta)
            else:
                # Add default=argparse.SUPPRESS so unprovided args don't appear in namespace
                parser.add_argument(f"--{entry_key}", default=argparse.SUPPRESS, help=entry_meta.desc)

    def _is_dict_type(self, entry_key: str) -> bool:
        """Check if a registered key has dict category."""
        if entry_key not in self.meta_info:
            return False
        entry_meta = self.meta_info[entry_key]
        proclist = self.category_proclist_cache.get(entry_meta.category)
        if proclist is None:
            return False
        return proclist["cast"] == "map"

    def _collect_leaf_keys(self, nested_dict: dict, prefix: str = "") -> list[str]:
        """
        Recursively collect all leaf keys from a nested dict as dot-notation strings.
        
        Stops recursion if the current key is registered as a dict type (its children
        are expected to be arbitrary and should not trigger unknown key warnings).
        """
        keys = []
        for k, v in nested_dict.items():
            full_key = f"{prefix}.{k}" if prefix else k
            # If this key is registered as dict type, don't recurse into its children
            if self._is_dict_type(full_key):
                keys.append(full_key)
            elif isinstance(v, dict):
                keys.extend(self._collect_leaf_keys(v, full_key))
            else:
                keys.append(full_key)
        return keys

    def _warn_unknown_keys(self, source_dict: dict, source_name: str) -> list[str]:
        """
        Warn about keys in source_dict that are not registered in meta_info.
        
        Returns the list of unknown keys (for testing purposes).
        """
        all_keys = self._collect_leaf_keys(source_dict)
        registered_keys = set(self.meta_info.keys())
        unknown_keys = [k for k in all_keys if k not in registered_keys]
        if unknown_keys:
            logger.warning(f"Unknown keys in {source_name} will be ignored: {unknown_keys}")
        return unknown_keys

    def _apply_config_file(self, config_tree, config_filepath, entry_config, warn_unknown_key: bool = False):
        """Load and apply a single config file."""
        config_ext = os.path.splitext(config_filepath)[1]
        if config_ext in [".yml", ".yaml"]:
            config_blob = load_yaml(config_filepath)
        else:
            raise RuntimeError(f"unsupported config type! got {repr(config_ext)}")

        # Warn about unknown keys if enabled
        if warn_unknown_key:
            self._warn_unknown_keys(config_blob, f"config file '{config_filepath}'")

        for entry_key in entry_config:
            entry_meta = self.meta_info[entry_key]
            index_status, raw_res = index_key(config_blob, entry_key)
            if index_status:
                cast_res = cast_to_res(raw_res, self.category_proclist_cache[entry_meta.category])
                set_value(config_tree, entry_key, cast_res)

    def _apply_cmdline_args(self, config_tree, parser, argv_segment):
        """Parse and apply a single command-line argument segment."""
        # Parse this segment
        namespace = parser.parse_args(argv_segment)
        parsed_dict = vars(namespace)  # Convert to dict

        # Only process explicitly provided args (due to SUPPRESS)
        entry_cmdline = list(entry_key for entry_key, entry_meta in self.meta_info.items()
                             if entry_meta.source in (ConfigEntrySource.COMMANDLINE_ONLY,
                                                      ConfigEntrySource.COMMANDLINE_OVER_CONFIG))

        for entry_key in entry_cmdline:
            entry_meta = self.meta_info[entry_key]

            # Handle boolean type
            if entry_meta.category == bool:
                raw_res = handle_cmd_bool(namespace, entry_key, entry_meta.cmdpattern)
                if raw_res == ConfigEntryValueUnspecified:
                    continue
            else:
                # Check if this arg was provided in this segment
                if entry_key not in parsed_dict:
                    continue

                raw_res = parsed_dict[entry_key]
                # Handle seq/map type cmdpattern
                if isinstance(entry_meta.cmdpattern, ConfigEntryCommandlineSeqPattern):
                    raw_res = handle_cmd_seq(raw_res, entry_meta.cmdpattern)
                elif isinstance(entry_meta.cmdpattern, ConfigEntryCommandlineMapPattern):
                    raw_res = handle_cmd_map(raw_res, entry_meta.cmdpattern)

            # Type cast and set value
            cast_res = cast_to_res(raw_res, self.category_proclist_cache[entry_meta.category])
            set_value(config_tree, entry_key, cast_res)

    def _apply_override(self, config_tree, cfg_override: dict, warn_unknown_key: bool = False):
        """
        Apply config overrides from a nested dict.
        
        Example: {"model": {"lr": 0.01, "batch_size": 32}}
        
        Only keys registered in meta_info are processed (consistent with _apply_config_file).
        """
        # Warn about unknown keys if enabled
        if warn_unknown_key:
            self._warn_unknown_keys(cfg_override, "cfg_override")

        for entry_key, entry_meta in self.meta_info.items():
            index_status, raw_value = index_key(cfg_override, entry_key)
            if index_status:
                cast_res = cast_to_res(raw_value, self.category_proclist_cache[entry_meta.category])
                set_value(config_tree, entry_key, cast_res)

    def _process_callbacks(self, config_tree):
        """Process all callbacks."""
        entry_callback_map = {}
        for entry_key, entry_meta in self.meta_info.items():
            if entry_meta.callback is not None and (entry_meta.callback.always or index_key(config_tree, entry_key)[1]
                                                    == ConfigEntryValueUnspecified):
                entry_callback_map[entry_key] = entry_meta.callback

        callback_run_order = resolve_callback_dependency(entry_callback_map)
        if callback_run_order is None:
            raise RuntimeError(f"circular dependency when solving callback order!")

        # process callback
        for entry_key in callback_run_order:
            entry_callback: ConfigEntryCallback = self.meta_info[entry_key].callback
            entry_value = index_key(config_tree, entry_key)[1]
            dep = {}
            for dep_key in entry_callback.dependency:
                dep[dep_key] = index_key(config_tree, dep_key)[1]
            callback_value = entry_callback(entry_key, entry_value, prog=self.prog, dep=dep)
            if callback_value != ConfigEntryValueUnspecified:
                set_value(config_tree, entry_key, callback_value)

    def _finalize_config(self, config_tree, strict):
        """Validate integrity and save config."""
        is_good, lack_key_list = check_config_integrity(self.meta_info_tree, config_tree)
        if strict and (not is_good):
            raise ValueError(f"unspecified value in config! key: {lack_key_list}")
        self.config = config_tree
        self.lack_key_list = lack_key_list

    def parse(self,
              parser: Optional[argparse.ArgumentParser] = None,
              arg_src=None,
              cfg_override=None,
              strict=True,
              warn_unknown_key: Optional[bool] = None):
        """
        Parse config files and command-line arguments in left-to-right order.

        :param parser: argparse.ArgumentParser
            parser to parse args from
        :param arg_src: list[str]
            arg list to parse
        :param cfg_override: dict
            config dict to override (nested format, e.g., {"model": {"lr": 0.01}})
        :param strict: bool
            if True, raise ValueError when required field has unspecified value in config
        :param warn_unknown_key: bool or None
            if True, log warning for unknown keys in config files and cfg_override;
            if None, use the instance attribute self.warn_unknown_key
        """
        # Resolve warn_unknown_key: per-call value overrides instance attribute
        _warn_unknown_key = self.warn_unknown_key if warn_unknown_key is None else warn_unknown_key

        # 1. Prepare default config
        _config = prepare_default_config(self.meta_info_tree)

        # 2. Get list of entries that can be read from config files
        entry_config = list(entry_key for entry_key, entry_meta in self.meta_info.items()
                            if entry_meta.source in (ConfigEntrySource.CONFIG_ONLY,
                                                     ConfigEntrySource.COMMANDLINE_OVER_CONFIG))

        # 3. Process pre-bound config files (bind_default_config_filepath)
        for config_filepath in self.bind_config_filepath_list:
            self._apply_config_file(_config, config_filepath, entry_config, warn_unknown_key=_warn_unknown_key)

        # 4. If no parser, skip command-line parsing
        if parser is None:
            # Still apply override if provided
            if cfg_override is not None:
                self._apply_override(_config, cfg_override, warn_unknown_key=_warn_unknown_key)
            self._process_callbacks(_config)
            self._finalize_config(_config, strict)
            return

        # 5. Segment-based argv processing
        if arg_src is None:
            arg_src = sys.argv[1:]

        segments = split_argv_by_cfg(arg_src)

        # 6. Process each segment in order
        for seg_type, seg_content in segments:
            if seg_type == 'cfg':
                # Load and apply config file
                self._apply_config_file(_config, seg_content, entry_config, warn_unknown_key=_warn_unknown_key)

            elif seg_type == 'args':
                # Parse and apply command-line arguments
                self._apply_cmdline_args(_config, parser, seg_content)

        # 7. Process override
        if cfg_override is not None:
            self._apply_override(_config, cfg_override, warn_unknown_key=_warn_unknown_key)

        # 8. Process callbacks
        self._process_callbacks(_config)

        # 9. Validate and save config
        self._finalize_config(_config, strict)

    def select(self, prefix: Optional[str] = None, strip=False):
        cfg = deepcopy(self.config)
        if strip:
            cfg = self.strip(cfg)
        if prefix is None:
            return cfg

        index_status, index_value = index_key(cfg, prefix)
        if not index_status:
            raise KeyError(f"prefix not found! got {prefix}")
        return index_value

    def strip(self, opt: Mapping):
        res = prepare_default_config(self.meta_info_tree)
        for entry_key in self.meta_info:
            index_status, index_value = index_key(opt, entry_key)
            if index_status and index_value != ConfigEntryValueUnspecified:
                set_value(res, entry_key, index_value)
            else:
                del_value(res, entry_key)
        return res

    def register_proxy(self, **kwarg):
        return RegisterProxy(self, **kwarg)


class RegisterProxy:

    def __init__(self, config_reg: ConfigRegistry, **kwarg) -> None:
        self.config_reg = config_reg
        self.kwarg = kwarg

    def register(self, *args, **kwargs):
        _kv = self.kwarg.copy()
        _kv.update(kwargs)
        self.config_reg.register(*args, **_kv)
        return self



"""Tests for cfg_override feature."""

import argparse
import pytest

from config_reg import ConfigRegistry, ConfigEntrySource
from config_reg.callback import ConfigEntryCallback


class TestCfgOverride:
    """Tests for cfg_override in parse()."""

    def test_override_with_nested_dict(self):
        reg = ConfigRegistry()
        reg.register("model.lr", category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)
        reg.register("model.batch_size", category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=16)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, [], cfg_override={"model": {"lr": 0.01, "batch_size": 32}})

        assert reg.config["model"]["lr"] == 0.01
        assert reg.config["model"]["batch_size"] == 32

    def test_override_deeply_nested(self):
        """Override should work with deeply nested structures."""
        reg = ConfigRegistry()
        reg.register("model.encoder.hidden_size",
                     category=int,
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     default=256)
        reg.register("model.encoder.num_layers",
                     category=int,
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     default=4)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, [], cfg_override={"model": {"encoder": {"hidden_size": 512, "num_layers": 8}}})

        assert reg.config["model"]["encoder"]["hidden_size"] == 512
        assert reg.config["model"]["encoder"]["num_layers"] == 8

    def test_override_type_casting(self):
        """Override values should be type-casted according to category."""
        reg = ConfigRegistry()
        reg.register("model.lr", category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)
        reg.register("model.batch_size", category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=16)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # Pass string values that need type casting
        reg.parse(parser, [], cfg_override={"model": {"lr": "0.01", "batch_size": "32"}})

        assert reg.config["model"]["lr"] == 0.01
        assert isinstance(reg.config["model"]["lr"], float)
        assert reg.config["model"]["batch_size"] == 32
        assert isinstance(reg.config["model"]["batch_size"], int)

    def test_override_unregistered_key_ignored(self):
        """Unregistered keys in override should be silently ignored."""
        reg = ConfigRegistry()
        reg.register("model.lr", category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # Include an unregistered key
        reg.parse(parser, [], cfg_override={"model": {"lr": 0.01, "unknown": 123}})

        assert reg.config["model"]["lr"] == 0.01
        assert "unknown" not in reg.config["model"]

    def test_warn_unknown_key_disabled_by_default(self, caplog):
        """By default, unknown keys should not trigger a warning."""
        import logging
        reg = ConfigRegistry()
        reg.register("model.lr", category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        with caplog.at_level(logging.WARNING):
            reg.parse(parser, [], cfg_override={"model": {"lr": 0.01, "unknown": 123}})

        assert "Unknown keys" not in caplog.text

    def test_warn_unknown_key_via_init(self, caplog):
        """Unknown keys should trigger warning when warn_unknown_key=True in init."""
        import logging
        reg = ConfigRegistry(warn_unknown_key=True)
        reg.register("model.lr", category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        with caplog.at_level(logging.WARNING):
            reg.parse(parser, [], cfg_override={"model": {"lr": 0.01, "unknown": 123}})

        assert "Unknown keys" in caplog.text
        assert "model.unknown" in caplog.text

    def test_warn_unknown_key_via_parse(self, caplog):
        """Unknown keys should trigger warning when warn_unknown_key=True in parse()."""
        import logging
        reg = ConfigRegistry()  # Default is False
        reg.register("model.lr", category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        with caplog.at_level(logging.WARNING):
            reg.parse(parser, [], cfg_override={"model": {"lr": 0.01, "unknown": 123}}, warn_unknown_key=True)

        assert "Unknown keys" in caplog.text
        assert "model.unknown" in caplog.text

    def test_warn_unknown_key_parse_overrides_init(self, caplog):
        """Per-call warn_unknown_key should override instance attribute."""
        import logging
        reg = ConfigRegistry(warn_unknown_key=True)  # Instance default is True
        reg.register("model.lr", category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # Override to False in parse()
        with caplog.at_level(logging.WARNING):
            reg.parse(parser, [], cfg_override={"model": {"lr": 0.01, "unknown": 123}}, warn_unknown_key=False)

        assert "Unknown keys" not in caplog.text

    def test_warn_unknown_key_dict_type_children_not_warned(self, caplog):
        """Children of dict-type registered keys should not trigger warnings."""
        import logging
        from config_reg.type_def import ConfigEntryCommandlineMapPattern

        reg = ConfigRegistry(warn_unknown_key=True)
        reg.register("model.lr", category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)
        # Register a dict-type key
        reg.register("model.params",
                     category=dict[str, int],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineMapPattern.COMMA_SEP_COLON)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        with caplog.at_level(logging.WARNING):
            # model.params children (any_key, another_key) should not be warned
            # model.unknown should be warned
            reg.parse(parser, [],
                      cfg_override={"model": {
                          "lr": 0.01,
                          "params": {
                              "any_key": 1,
                              "another_key": 2
                          },
                          "unknown": 123
                      }})

        assert "model.unknown" in caplog.text
        assert "model.params.any_key" not in caplog.text
        assert "model.params.another_key" not in caplog.text

    def test_override_ignores_source_constraint(self):
        """Override should work regardless of source constraint."""
        reg = ConfigRegistry()
        # CONFIG_ONLY normally can't be set via commandline, but override should work
        reg.register("model.lr", category=float, source=ConfigEntrySource.CONFIG_ONLY, default=0.001)
        # COMMANDLINE_ONLY normally can't be set via config file, but override should work
        reg.register("model.batch_size", category=int, source=ConfigEntrySource.COMMANDLINE_ONLY, default=16)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, [], cfg_override={"model": {"lr": 0.01, "batch_size": 32}})

        assert reg.config["model"]["lr"] == 0.01
        assert reg.config["model"]["batch_size"] == 32

    def test_override_after_config_file(self, tmp_path):
        """Override should be applied after config file, overriding its values."""
        # Create a temp config file
        config_file = tmp_path / "config.yaml"
        config_file.write_text("model:\n  lr: 0.005\n  batch_size: 64\n")

        reg = ConfigRegistry()
        reg.register("model.lr", category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)
        reg.register("model.batch_size", category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=16)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # Parse with config file, then override lr
        reg.parse(parser, ["-c", str(config_file)], cfg_override={"model": {"lr": 0.01}})

        assert reg.config["model"]["lr"] == 0.01  # Overridden
        assert reg.config["model"]["batch_size"] == 64  # From config file

    def test_override_after_cmdline(self):
        """Override should be applied after cmdline args, overriding them."""
        reg = ConfigRegistry()
        reg.register("model.lr", category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)
        reg.register("model.batch_size", category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=16)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # Parse with cmdline args, then override lr
        reg.parse(parser, ["--model.lr", "0.005", "--model.batch_size", "64"], cfg_override={"model": {"lr": 0.01}})

        assert reg.config["model"]["lr"] == 0.01  # Overridden
        assert reg.config["model"]["batch_size"] == 64  # From cmdline

    def test_override_partial(self):
        """Override should only affect specified keys, leaving others unchanged."""
        reg = ConfigRegistry()
        reg.register("model.lr", category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)
        reg.register("model.batch_size", category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=16)
        reg.register("model.epochs", category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=100)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # Only override lr, batch_size and epochs should keep defaults
        reg.parse(parser, [], cfg_override={"model": {"lr": 0.01}})

        assert reg.config["model"]["lr"] == 0.01
        assert reg.config["model"]["batch_size"] == 16  # Default
        assert reg.config["model"]["epochs"] == 100  # Default


class TestCfgOverrideWithCallback:
    """Tests for cfg_override interaction with callbacks."""

    def test_callback_receives_overridden_value(self):
        """Callbacks should receive the overridden value in their dependencies."""
        received_values = {}

        class RecordingCallback(ConfigEntryCallback):
            dependency = ["model.lr"]
            always = True

            def __call__(self, curr_key, curr_value, prog, dep):
                received_values["lr"] = dep["model.lr"]
                return curr_value if curr_value is not None else 0

        reg = ConfigRegistry()
        reg.register("model.lr", category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)
        reg.register("model.computed", category=float, source=ConfigEntrySource.CALLBACK, callback=RecordingCallback())

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, [], cfg_override={"model": {"lr": 0.01}})

        assert received_values["lr"] == 0.01

    def test_callback_computes_from_overridden_value(self):
        """Callbacks should compute based on overridden values."""

        class DoubleCallback(ConfigEntryCallback):
            dependency = ["model.lr"]
            always = True

            def __call__(self, curr_key, curr_value, prog, dep):
                return dep["model.lr"] * 2

        reg = ConfigRegistry()
        reg.register("model.lr", category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)
        reg.register("model.lr_doubled", category=float, source=ConfigEntrySource.CALLBACK, callback=DoubleCallback())

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, [], cfg_override={"model": {"lr": 0.01}})

        assert reg.config["model"]["lr_doubled"] == 0.02

    def test_override_does_not_affect_callback_entry_directly(self):
        """Override on a CALLBACK source entry is applied but callback runs after."""

        class FixedCallback(ConfigEntryCallback):
            dependency = []
            always = True

            def __call__(self, curr_key, curr_value, prog, dep):
                return 999.0

        reg = ConfigRegistry()
        reg.register("model.computed", category=float, source=ConfigEntrySource.CALLBACK, callback=FixedCallback())

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # Even if we try to override a CALLBACK entry, the callback runs after and sets the final value
        reg.parse(parser, [], cfg_override={"model": {"computed": 123.0}})

        # Callback runs after override, so callback's return value is the final value
        assert reg.config["model"]["computed"] == 999.0


class TestCfgOverrideNoParser:
    """Tests for cfg_override when parser is None."""

    def test_override_without_parser(self):
        """Override should work even without a parser."""
        reg = ConfigRegistry()
        reg.register("model.lr", category=float, source=ConfigEntrySource.CONFIG_ONLY, default=0.001)
        reg.register("model.batch_size", category=int, source=ConfigEntrySource.CONFIG_ONLY, default=16)

        reg.parse(None, cfg_override={"model": {"lr": 0.01, "batch_size": 32}})

        assert reg.config["model"]["lr"] == 0.01
        assert reg.config["model"]["batch_size"] == 32



"""Tests for instantiate functionality in convert.py"""

import functools
import pytest
from config_reg.convert import instantiate, instantiate_node, InstantiationException


# Test helper classes
class SimpleClass:

    def __init__(self, value):
        self.value = value


class ParentClass:

    def __init__(self, child, name="parent"):
        self.child = child
        self.name = name


class ChildClass:

    def __init__(self, data):
        self.data = data


class GrandchildClass:

    def __init__(self, id):
        self.id = id


# =============================================================================
# Basic Instantiation Tests
# =============================================================================


class TestBasicInstantiation:

    def test_simple_instantiation(self):
        """Test basic instantiation with _target_"""
        config = {"_target_": "test.test_instantiate.SimpleClass", "value": 42}
        result = instantiate(config)
        assert isinstance(result, SimpleClass)
        assert result.value == 42

    def test_instantiation_with_callable_target(self):
        """Test instantiation with callable (not string) as _target_"""
        config = {"_target_": SimpleClass, "value": "hello"}
        result = instantiate(config)
        assert isinstance(result, SimpleClass)
        assert result.value == "hello"

    def test_nested_instantiation(self):
        """Test recursive instantiation of nested configs"""
        config = {
            "_target_": "test.test_instantiate.ParentClass",
            "child": {
                "_target_": "test.test_instantiate.ChildClass",
                "data": {
                    "key": "value"
                }
            },
            "name": "test_parent"
        }
        result = instantiate(config)
        assert isinstance(result, ParentClass)
        assert isinstance(result.child, ChildClass)
        assert result.child.data == {"key": "value"}

    def test_none_config_returns_none(self):
        """Test that None config returns None"""
        result = instantiate_node(None)
        assert result is None

    def test_primitive_config_returns_as_is(self):
        """Test that primitives pass through unchanged"""
        assert instantiate_node(42) == 42
        assert instantiate_node("hello") == "hello"
        assert instantiate_node(3.14) == 3.14


# =============================================================================
# Node-Level _recursive_ Override Tests
# =============================================================================


class TestNodeLevelRecursive:

    def test_top_level_recursive_false(self):
        """Test _recursive_: false at top level stops all recursion"""
        config = {
            "_target_": "test.test_instantiate.ParentClass",
            "_recursive_": False,
            "child": {
                "_target_": "test.test_instantiate.ChildClass",
                "data": "test"
            },
            "name": "parent"
        }
        result = instantiate(config)
        assert isinstance(result, ParentClass)
        # child should remain as dict, not instantiated
        assert isinstance(result.child, dict)
        assert result.child["_target_"] == "test.test_instantiate.ChildClass"

    def test_node_level_recursive_false_stops_children(self):
        """Test _recursive_: false at node level stops that node's children only"""
        config = {
            "_target_": "test.test_instantiate.ParentClass",
            "_recursive_": True,  # top level recursive
            "child": {
                "_target_": "test.test_instantiate.ParentClass",
                "_recursive_": False,  # stop recursion here
                "child": {
                    "_target_": "test.test_instantiate.ChildClass",
                    "data": "should_be_dict"
                },
                "name": "middle"
            },
            "name": "top"
        }
        result = instantiate(config)

        # Top level instantiated
        assert isinstance(result, ParentClass)
        assert result.name == "top"

        # Middle level instantiated (because parent's _recursive_ is True)
        assert isinstance(result.child, ParentClass)
        assert result.child.name == "middle"

        # Grandchild should remain as dict (because middle's _recursive_ is False)
        assert isinstance(result.child.child, dict)
        assert result.child.child["_target_"] == "test.test_instantiate.ChildClass"

    def test_deep_nested_recursive_override(self):
        """Test _recursive_ override at multiple levels"""
        config = {
            "_target_": "test.test_instantiate.ParentClass",
            "child": {
                "_target_": "test.test_instantiate.ParentClass",
                "_recursive_": False,  # stop here
                "child": {
                    "_target_": "test.test_instantiate.ParentClass",
                    "_recursive_": True,  # this won't matter, parent stopped
                    "child": {
                        "_target_": "test.test_instantiate.ChildClass",
                        "data": "deep"
                    }
                }
            }
        }
        result = instantiate(config)

        assert isinstance(result, ParentClass)
        assert isinstance(result.child, ParentClass)
        # The third level should be dict because middle set _recursive_: False
        assert isinstance(result.child.child, dict)


# =============================================================================
# Node-Level _partial_ Override Tests
# =============================================================================


class TestNodeLevelPartial:

    def test_top_level_partial(self):
        """Test _partial_: true creates functools.partial"""
        config = {"_target_": "test.test_instantiate.SimpleClass", "_partial_": True, "value": 100}
        result = instantiate(config)
        assert isinstance(result, functools.partial)
        # Call the partial to get the actual instance
        instance = result()
        assert isinstance(instance, SimpleClass)
        assert instance.value == 100

    def test_node_level_partial_override(self):
        """Test _partial_ at node level creates partial for that node only"""
        config = {
            "_target_": "test.test_instantiate.ParentClass",
            "_partial_": False,  # top level fully instantiated
            "child": {
                "_target_": "test.test_instantiate.SimpleClass",
                "_partial_": True,  # child should be partial
                "value": 50
            },
            "name": "parent"
        }
        result = instantiate(config)

        # Parent fully instantiated
        assert isinstance(result, ParentClass)
        # Child should be a partial
        assert isinstance(result.child, functools.partial)
        # Calling the partial creates the instance
        child_instance = result.child()
        assert isinstance(child_instance, SimpleClass)
        assert child_instance.value == 50


# =============================================================================
# Error Handling Tests
# =============================================================================


class TestErrorHandling:

    def test_missing_target_raises(self):
        """Test that missing _target_ raises InstantiationException"""
        config = {"value": 42}
        with pytest.raises(InstantiationException):
            instantiate(config)

    def test_empty_target_raises(self):
        """Test that empty _target_ raises InstantiationException"""
        config = {"_target_": "", "value": 42}
        with pytest.raises(InstantiationException):
            instantiate(config)

    def test_none_target_raises(self):
        """Test that None _target_ raises InstantiationException"""
        config = {"_target_": None, "value": 42}
        with pytest.raises(InstantiationException):
            instantiate(config)

    def test_invalid_target_path_raises(self):
        """Test that invalid target path raises InstantiationException"""
        config = {"_target_": "nonexistent.module.Class", "value": 42}
        with pytest.raises(InstantiationException):
            instantiate(config)

    def test_invalid_recursive_type_raises(self):
        """Test that non-bool _recursive_ raises TypeError"""
        config = {
            "_target_": "test.test_instantiate.SimpleClass",
            "_recursive_": "yes",  # should be bool
            "value": 42
        }
        with pytest.raises(TypeError):
            instantiate(config)

    def test_invalid_partial_type_raises(self):
        """Test that non-bool _partial_ raises TypeError"""
        config = {
            "_target_": "test.test_instantiate.SimpleClass",
            "_partial_": 1,  # should be bool
            "value": 42
        }
        with pytest.raises(TypeError):
            instantiate(config)


# =============================================================================
# List Config Tests
# =============================================================================


class TestListConfig:

    def test_list_of_targets(self):
        """Test instantiation of list containing target configs"""
        config = {
            "_target_": "test.test_instantiate.ParentClass",
            "child": [
                {
                    "_target_": "test.test_instantiate.SimpleClass",
                    "value": 1
                },
                {
                    "_target_": "test.test_instantiate.SimpleClass",
                    "value": 2
                },
            ],
            "name": "list_parent"
        }
        result = instantiate(config)
        assert isinstance(result, ParentClass)
        assert isinstance(result.child, list)
        assert len(result.child) == 2
        assert all(isinstance(c, SimpleClass) for c in result.child)
        assert result.child[0].value == 1
        assert result.child[1].value == 2


# =============================================================================
# _args_ Tests
# =============================================================================


class TestPositionalArgs:

    def test_args_in_config(self):
        """Test _args_ for positional arguments"""
        config = {"_target_": "test.test_instantiate.SimpleClass", "_args_": ["positional_value"]}
        result = instantiate(config)
        assert isinstance(result, SimpleClass)
        assert result.value == "positional_value"


# =============================================================================
# Dataclass Config Tests
# =============================================================================

from dataclasses import dataclass, field


# Dataclass test helper classes
@dataclass
class SimpleDataclassConfig:
    _target_: str = "test.test_instantiate.SimpleClass"
    value: int = 42


@dataclass
class ChildDataclassConfig:
    _target_: str = "test.test_instantiate.ChildClass"
    data: str = "default_data"


@dataclass
class ParentDataclassConfig:
    _target_: str = "test.test_instantiate.ParentClass"
    child: ChildDataclassConfig = field(default_factory=ChildDataclassConfig)
    name: str = "dataclass_parent"


@dataclass
class RecursiveFalseParentConfig:
    _target_: str = "test.test_instantiate.ParentClass"
    _recursive_: bool = False
    child: ChildDataclassConfig = field(default_factory=ChildDataclassConfig)
    name: str = "non_recursive_parent"


@dataclass
class PartialChildConfig:
    _target_: str = "test.test_instantiate.SimpleClass"
    _partial_: bool = True
    value: int = 100


@dataclass
class ParentWithPartialChildConfig:
    _target_: str = "test.test_instantiate.ParentClass"
    child: PartialChildConfig = field(default_factory=PartialChildConfig)
    name: str = "parent_with_partial"


@dataclass
class CallableTargetConfig:
    _target_: type = SimpleClass  # Callable, not string
    value: str = "callable_target"


class DerivedValueClass:
    """Helper class that accepts base_value, multiplier, and value"""
    def __init__(self, base_value, multiplier, value):
        self.base_value = base_value
        self.multiplier = multiplier
        self.value = value


@dataclass
class DerivedFieldConfig:
    """Config where value is derived from other fields via __post_init__"""
    _target_: str = "test.test_instantiate.DerivedValueClass"
    base_value: int = 10
    multiplier: int = 5
    value: int = field(init=False)  # Derived field - will be computed

    def __post_init__(self):
        self.value = self.base_value * self.multiplier


@dataclass
class GrandchildDataclassConfig:
    _target_: str = "test.test_instantiate.GrandchildClass"
    id: int = 1


@dataclass
class MiddleDataclassConfig:
    _target_: str = "test.test_instantiate.ChildClass"
    _recursive_: bool = False  # Stop recursion here
    data: GrandchildDataclassConfig = field(default_factory=GrandchildDataclassConfig)


@dataclass
class TopDataclassConfig:
    _target_: str = "test.test_instantiate.ParentClass"
    child: MiddleDataclassConfig = field(default_factory=MiddleDataclassConfig)
    name: str = "top"


class TestDataclassConfig:

    def test_simple_dataclass(self):
        """Test basic dataclass config instantiation"""
        cfg = SimpleDataclassConfig(value=100)
        result = instantiate(cfg)
        assert isinstance(result, SimpleClass)
        assert result.value == 100

    def test_nested_dataclass(self):
        """Test nested dataclass configs are recursively instantiated"""
        cfg = ParentDataclassConfig(
            child=ChildDataclassConfig(data="nested_value"),
            name="test_parent"
        )
        result = instantiate(cfg)
        assert isinstance(result, ParentClass)
        assert isinstance(result.child, ChildClass)
        assert result.child.data == "nested_value"
        assert result.name == "test_parent"

    def test_dataclass_recursive_false_stops_children(self):
        """Test _recursive_: False in dataclass stops child instantiation"""
        cfg = RecursiveFalseParentConfig(
            child=ChildDataclassConfig(data="should_be_dict")
        )
        result = instantiate(cfg)
        assert isinstance(result, ParentClass)
        # child should remain as dict, not instantiated
        assert isinstance(result.child, dict)
        assert result.child["_target_"] == "test.test_instantiate.ChildClass"
        assert result.child["data"] == "should_be_dict"

    def test_dataclass_node_level_recursive_false(self):
        """Test node-level _recursive_: False in nested dataclass"""
        cfg = TopDataclassConfig()
        result = instantiate(cfg)
        
        # Top level instantiated
        assert isinstance(result, ParentClass)
        assert result.name == "top"
        
        # Middle level instantiated (top's _recursive_ is True by default)
        assert isinstance(result.child, ChildClass)
        
        # Grandchild should be dict (middle's _recursive_ is False)
        assert isinstance(result.child.data, dict)
        assert result.child.data["_target_"] == "test.test_instantiate.GrandchildClass"

    def test_dataclass_partial_child(self):
        """Test _partial_: True in nested dataclass creates functools.partial"""
        cfg = ParentWithPartialChildConfig()
        result = instantiate(cfg)
        
        assert isinstance(result, ParentClass)
        # child should be a partial
        assert isinstance(result.child, functools.partial)
        # Calling the partial creates the instance
        child_instance = result.child()
        assert isinstance(child_instance, SimpleClass)
        assert child_instance.value == 100

    def test_dataclass_callable_target(self):
        """Test dataclass with callable (type) as _target_"""
        cfg = CallableTargetConfig(value="test_callable")
        result = instantiate(cfg)
        assert isinstance(result, SimpleClass)
        assert result.value == "test_callable"

    def test_dataclass_derived_field(self):
        """Test dataclass with derived field (init=False) via __post_init__"""
        cfg = DerivedFieldConfig(base_value=20, multiplier=3)
        # value should be 20 * 3 = 60
        assert cfg.value == 60
        result = instantiate(cfg)
        assert isinstance(result, DerivedValueClass)
        # The derived value should be passed to target
        assert result.value == 60
        assert result.base_value == 20
        assert result.multiplier == 3
        assert result.value == 60

    def test_dataclass_mixed_with_dict(self):
        """Test dataclass containing dict children"""
        @dataclass
        class MixedConfig:
            _target_: str = "test.test_instantiate.ParentClass"
            child: dict = field(default_factory=lambda: {
                "_target_": "test.test_instantiate.ChildClass",
                "data": "from_dict"
            })
            name: str = "mixed"

        cfg = MixedConfig()
        result = instantiate(cfg)
        assert isinstance(result, ParentClass)
        assert isinstance(result.child, ChildClass)
        assert result.child.data == "from_dict"

    def test_dict_containing_dataclass(self):
        """Test dict config containing dataclass children"""
        config = {
            "_target_": "test.test_instantiate.ParentClass",
            "child": ChildDataclassConfig(data="dataclass_in_dict"),
            "name": "dict_parent"
        }
        result = instantiate(config)
        assert isinstance(result, ParentClass)
        assert isinstance(result.child, ChildClass)
        assert result.child.data == "dataclass_in_dict"

    def test_dataclass_list_children(self):
        """Test dataclass with list of dataclass children"""
        @dataclass
        class ListChildrenConfig:
            _target_: str = "test.test_instantiate.ParentClass"
            child: list = field(default_factory=lambda: [
                SimpleDataclassConfig(value=1),
                SimpleDataclassConfig(value=2),
            ])
            name: str = "list_parent"

        cfg = ListChildrenConfig()
        result = instantiate(cfg)
        assert isinstance(result, ParentClass)
        assert isinstance(result.child, list)
        assert len(result.child) == 2
        assert all(isinstance(c, SimpleClass) for c in result.child)
        assert result.child[0].value == 1
        assert result.child[1].value == 2

    def test_dataclass_default_values(self):
        """Test dataclass with default values"""
        cfg = SimpleDataclassConfig()  # Use all defaults
        result = instantiate(cfg)
        assert isinstance(result, SimpleClass)
        assert result.value == 42  # default value

    def test_dataclass_deep_nesting(self):
        """Test deeply nested dataclass configs"""
        @dataclass
        class Level3Config:
            _target_: str = "test.test_instantiate.GrandchildClass"
            id: int = 3

        @dataclass
        class Level2Config:
            _target_: str = "test.test_instantiate.ChildClass"
            data: Level3Config = field(default_factory=Level3Config)

        @dataclass
        class Level1Config:
            _target_: str = "test.test_instantiate.ParentClass"
            child: Level2Config = field(default_factory=Level2Config)
            name: str = "level1"

        cfg = Level1Config()
        result = instantiate(cfg)
        
        assert isinstance(result, ParentClass)
        assert isinstance(result.child, ChildClass)
        assert isinstance(result.child.data, GrandchildClass)
        assert result.child.data.id == 3



import yaml


def load_yaml(filename):
    with open(filename, "r") as f:
        res = yaml.safe_load(f)
    return res


def save_yaml(res, filename):
    with open(filename, "w") as f:
        yaml.dump(res, f)



"""
Integration tests for ordered parsing of config files and command-line arguments
"""
import pytest
import argparse
import tempfile
import os
import yaml

from config_reg import ConfigRegistry
from config_reg.type_def import (
    ConfigEntrySource,
    ConfigEntryCommandlineBoolPattern,
    ConfigEntryValueUnspecified,
)


@pytest.fixture
def temp_config_dir():
    """Create a temporary directory for config files"""
    with tempfile.TemporaryDirectory() as tmpdir:
        yield tmpdir


def create_yaml_file(directory, filename, content):
    """Helper to create a yaml config file"""
    filepath = os.path.join(directory, filename)
    with open(filepath, 'w') as f:
        yaml.dump(content, f)
    return filepath


class TestOrderedParsing:
    """Tests for ordered parsing behavior"""

    def test_args_before_config(self, temp_config_dir):
        """Test: args set before config file should be overridden by config"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'a': 10, 'b': 20})

        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('b', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # --a 1 comes before config, so config should override it
        reg.parse(parser, ['--a', '1', '-c', cfg1_path], strict=False)

        assert reg.select()['a'] == 10  # cfg1 overrides --a 1
        assert reg.select()['b'] == 20

    def test_args_after_config(self, temp_config_dir):
        """Test: args set after config file should override config"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'a': 10, 'b': 20})

        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('b', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # config comes before --a 1, so --a 1 should override config
        reg.parse(parser, ['-c', cfg1_path, '--a', '1'], strict=False)

        assert reg.select()['a'] == 1  # --a 1 overrides cfg1
        assert reg.select()['b'] == 20

    def test_interleaved_args_and_config(self, temp_config_dir):
        """Test: args → config → args pattern"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'a': 10, 'b': 20, 'c': 30})

        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('b', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('c', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # --a 1 is before config (overridden), --b 200 is after config (overrides)
        reg.parse(parser, ['--a', '1', '-c', cfg1_path, '--b', '200'], strict=False)

        assert reg.select()['a'] == 10  # cfg1 overrides --a 1
        assert reg.select()['b'] == 200  # --b 200 overrides cfg1
        assert reg.select()['c'] == 30  # from cfg1

    def test_multiple_config_files(self, temp_config_dir):
        """Test: multiple config files override each other"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'a': 10, 'b': 20})
        cfg2_path = create_yaml_file(temp_config_dir, 'cfg2.yaml', {'a': 100})

        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('b', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '-c', cfg2_path], strict=False)

        assert reg.select()['a'] == 100  # cfg2 overrides cfg1
        assert reg.select()['b'] == 20  # from cfg1 (cfg2 doesn't have b)

    def test_config_between_args(self, temp_config_dir):
        """Test: config file between two arguments"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'a': 10, 'b': 20})

        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('b', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # Both args are after config, so both should override
        reg.parse(parser, ['-c', cfg1_path, '--a', '1', '--b', '2'], strict=False)

        assert reg.select()['a'] == 1
        assert reg.select()['b'] == 2

    def test_complex_interleaved(self, temp_config_dir):
        """Test: complex interleaving pattern"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'a': 10, 'b': 20, 'c': 30})
        cfg2_path = create_yaml_file(temp_config_dir, 'cfg2.yaml', {'b': 200, 'c': 300})

        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('b', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('c', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # --a 1 → cfg1 → --b 99 → cfg2 → --c 999
        reg.parse(parser, ['--a', '1', '-c', cfg1_path, '--b', '99', '--cfg', cfg2_path, '--c', '999'], strict=False)

        assert reg.select()['a'] == 10  # cfg1 overrides --a 1
        assert reg.select()['b'] == 200  # cfg2 overrides --b 99
        assert reg.select()['c'] == 999  # --c 999 overrides cfg2

    def test_cfg_equals_format(self, temp_config_dir):
        """Test: --cfg= format"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'a': 10})

        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, [f'--cfg={cfg1_path}'], strict=False)

        assert reg.select()['a'] == 10

    def test_only_args_no_config(self, temp_config_dir):
        """Test: only command-line args, no config files"""
        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('b', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['--a', '1', '--b', '2'], strict=False)

        assert reg.select()['a'] == 1
        assert reg.select()['b'] == 2

    def test_only_config_no_args(self, temp_config_dir):
        """Test: only config files, no command-line args"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'a': 10, 'b': 20})

        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('b', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path], strict=False)

        assert reg.select()['a'] == 10
        assert reg.select()['b'] == 20

    def test_empty_args(self, temp_config_dir):
        """Test: empty argument list"""
        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=5)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, [], strict=False)

        assert reg.select()['a'] == 5  # default value


class TestBooleanOrdering:
    """Tests for boolean arguments with ordered parsing"""

    def test_bool_set_true_after_config(self, temp_config_dir):
        """Test: SET_TRUE bool after config"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'flag': False})

        reg = ConfigRegistry()
        reg.register('flag',
                     category=bool,
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineBoolPattern.SET_TRUE)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '--flag'], strict=False)

        assert reg.select()['flag'] is True  # --flag overrides config

    def test_bool_set_true_before_config(self, temp_config_dir):
        """Test: SET_TRUE bool before config"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'flag': False})

        reg = ConfigRegistry()
        reg.register('flag',
                     category=bool,
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineBoolPattern.SET_TRUE)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['--flag', '-c', cfg1_path], strict=False)

        assert reg.select()['flag'] is False  # config overrides --flag

    def test_bool_on_off_pattern(self, temp_config_dir):
        """Test: ON_OFF bool pattern"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'flag': True})

        reg = ConfigRegistry()
        reg.register('flag',
                     category=bool,
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineBoolPattern.ON_OFF)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '--flag__off'], strict=False)

        assert reg.select()['flag'] is False  # --flag__off overrides config

    def test_bool_set_false_after_config(self, temp_config_dir):
        """Test: SET_FALSE bool after config"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'flag': True})

        reg = ConfigRegistry()
        reg.register('flag',
                     category=bool,
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineBoolPattern.SET_FALSE)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '--flag'], strict=False)

        assert reg.select()['flag'] is False  # --flag sets to False, overrides config

    def test_bool_set_false_before_config(self, temp_config_dir):
        """Test: SET_FALSE bool before config"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'flag': False})

        reg = ConfigRegistry()
        reg.register('flag',
                     category=bool,
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineBoolPattern.SET_FALSE)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # --flag sets to False, but config (also False) comes after
        reg.parse(parser, ['--flag', '-c', cfg1_path], strict=False)

        assert reg.select()['flag'] is False  # config overrides --flag

    def test_bool_on_off_enable(self, temp_config_dir):
        """Test: ON_OFF pattern with --flag (sets to True)"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'flag': False})

        reg = ConfigRegistry()
        reg.register('flag',
                     category=bool,
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineBoolPattern.ON_OFF)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '--flag'], strict=False)

        assert reg.select()['flag'] is True  # --flag enables, overrides config

    def test_bool_with_default_no_config(self, temp_config_dir):
        """Test: bool with default, no config file"""
        reg = ConfigRegistry()
        reg.register('flag',
                     category=bool,
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     default=True,
                     cmdpattern=ConfigEntryCommandlineBoolPattern.SET_TRUE)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, [], strict=False)

        assert reg.select()['flag'] is True  # default value

    def test_bool_unspecified_keeps_config(self, temp_config_dir):
        """Test: bool not specified on cmdline keeps config value"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'flag': True, 'other': 10})

        reg = ConfigRegistry()
        reg.register('flag',
                     category=bool,
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineBoolPattern.SET_TRUE)
        reg.register('other', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # Don't specify --flag, only --other
        reg.parse(parser, ['-c', cfg1_path, '--other', '99'], strict=False)

        assert reg.select()['flag'] is True  # keeps config value
        assert reg.select()['other'] == 99

    def test_bool_multiple_configs_override(self, temp_config_dir):
        """Test: later config overrides earlier bool value"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'flag': True})
        cfg2_path = create_yaml_file(temp_config_dir, 'cfg2.yaml', {'flag': False})

        reg = ConfigRegistry()
        reg.register('flag',
                     category=bool,
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineBoolPattern.SET_TRUE)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '-c', cfg2_path], strict=False)

        assert reg.select()['flag'] is False  # cfg2 overrides cfg1

    def test_bool_on_off_no_default_no_flag(self):
        """Test: ON_OFF pattern without default, no flag provided - returns Unspecified"""
        reg = ConfigRegistry()
        reg.register(
            'flag',
            category=bool,
            source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
            required=True,  # Mark as required
            cmdpattern=ConfigEntryCommandlineBoolPattern.ON_OFF)
        # No default is set

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # Neither --flag nor --flag__off is provided, no config file
        # With strict=True and required=True, this should raise ValueError
        with pytest.raises(ValueError, match="unspecified value in config"):
            reg.parse(parser, [], strict=True)

    def test_bool_on_off_no_default_no_flag_non_strict(self):
        """Test: ON_OFF pattern without default, no flag, non-strict mode"""
        reg = ConfigRegistry()
        reg.register('flag',
                     category=bool,
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineBoolPattern.ON_OFF)
        # No default is set

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # Neither --flag nor --flag__off is provided, no config file
        # With strict=False, should complete but value is Unspecified
        reg.parse(parser, [], strict=False)

        result = reg.select()
        # The key exists in result, but value is ConfigEntryValueUnspecified
        assert 'flag' in result
        assert result['flag'] is ConfigEntryValueUnspecified

        # After strip, the key should be removed
        result_stripped = reg.select(strip=True)
        assert 'flag' not in result_stripped


class TestNestedKeys:
    """Tests for nested config keys"""

    def test_nested_key_override(self, temp_config_dir):
        """Test: nested keys like model.lr"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'model': {'lr': 0.001, 'batch_size': 32}})

        reg = ConfigRegistry()
        reg.register('model.lr', category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('model.batch_size', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '--model.lr', '0.01'], strict=False)

        assert reg.select()['model']['lr'] == 0.01
        assert reg.select()['model']['batch_size'] == 32

    def test_deep_nested_keys(self, temp_config_dir):
        """Test: deep nesting like a.b.c.d"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'a': {'b': {'c': {'d': 1, 'e': 2}}}})

        reg = ConfigRegistry()
        reg.register('a.b.c.d', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('a.b.c.e', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '--a.b.c.d', '100'], strict=False)

        assert reg.select()['a']['b']['c']['d'] == 100
        assert reg.select()['a']['b']['c']['e'] == 2

    def test_nested_partial_override_multiple_configs(self, temp_config_dir):
        """Test: partial override with multiple config files"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml',
                                     {'model': {
                                         'lr': 0.001,
                                         'batch_size': 32,
                                         'hidden': 256
                                     }})
        cfg2_path = create_yaml_file(
            temp_config_dir,
            'cfg2.yaml',
            {
                'model': {
                    'lr': 0.01
                }  # only override lr
            })

        reg = ConfigRegistry()
        reg.register('model.lr', category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('model.batch_size', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('model.hidden', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '-c', cfg2_path], strict=False)

        assert reg.select()['model']['lr'] == 0.01  # from cfg2
        assert reg.select()['model']['batch_size'] == 32  # from cfg1
        assert reg.select()['model']['hidden'] == 256  # from cfg1

    def test_nested_cmdline_only(self, temp_config_dir):
        """Test: nested keys with COMMANDLINE_ONLY source"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'train': {'epochs': 100, 'verbose': True}})

        reg = ConfigRegistry()
        reg.register('train.epochs', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)
        reg.register('train.verbose',
                     category=bool,
                     source=ConfigEntrySource.COMMANDLINE_ONLY,
                     cmdpattern=ConfigEntryCommandlineBoolPattern.SET_TRUE)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # train.verbose from config should be ignored (COMMANDLINE_ONLY)
        reg.parse(parser, ['-c', cfg1_path, '--train.verbose'], strict=False)

        assert reg.select()['train']['epochs'] == 100
        assert reg.select()['train']['verbose'] is True  # from cmdline, not config

    def test_nested_with_defaults(self, temp_config_dir):
        """Test: nested keys with default values"""
        reg = ConfigRegistry()
        reg.register('model.lr', category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=0.001)
        reg.register('model.batch_size', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG, default=32)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['--model.batch_size', '64'], strict=False)

        assert reg.select()['model']['lr'] == 0.001  # default
        assert reg.select()['model']['batch_size'] == 64  # cmdline

    def test_nested_config_before_cmdline(self, temp_config_dir):
        """Test: nested key cmdline before config gets overridden"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'opt': {'learning_rate': 0.1}})

        reg = ConfigRegistry()
        reg.register('opt.learning_rate', category=float, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # cmdline before config, so config should win
        reg.parse(parser, ['--opt.learning_rate', '0.001', '-c', cfg1_path], strict=False)

        assert reg.select()['opt']['learning_rate'] == 0.1  # config overrides


class TestBindDefaultConfig:
    """Tests for bind_default_config_filepath"""

    def test_bind_default_before_cmdline(self, temp_config_dir):
        """Test: bind_default_config_filepath is applied before cmdline parsing"""
        cfg_default = create_yaml_file(temp_config_dir, 'default.yaml', {'a': 5})
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'a': 10})

        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        # Bind default config
        reg.bind_default_config_filepath(cfg_default)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # Both configs should be applied, but cfg1 comes after
        reg.parse(parser, ['-c', cfg1_path], strict=False)

        assert reg.select()['a'] == 10  # cfg1 overrides default

    def test_bind_default_with_cmdline_override(self, temp_config_dir):
        """Test: cmdline args override bind_default config"""
        cfg_default = create_yaml_file(temp_config_dir, 'default.yaml', {'a': 5})

        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        reg.bind_default_config_filepath(cfg_default)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['--a', '100'], strict=False)

        assert reg.select()['a'] == 100  # --a overrides default config


class TestConfigOnlyEntries:
    """Tests for CONFIG_ONLY entries"""

    def test_config_only_not_from_cmdline(self, temp_config_dir):
        """Test: CONFIG_ONLY entries are only set from config files"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'a': 10, 'b': 20})

        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.CONFIG_ONLY)
        reg.register('b', category=int, source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        # --a should not be registered (CONFIG_ONLY), only --b
        reg.parse(parser, ['-c', cfg1_path, '--b', '99'], strict=False)

        assert reg.select()['a'] == 10  # from config
        assert reg.select()['b'] == 99  # from cmdline


class TestCommandlineOnlyEntries:
    """Tests for COMMANDLINE_ONLY entries"""

    def test_commandline_only_not_from_config(self, temp_config_dir):
        """Test: COMMANDLINE_ONLY entries ignore config files"""
        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'a': 10})

        reg = ConfigRegistry()
        reg.register('a', category=int, source=ConfigEntrySource.COMMANDLINE_ONLY)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '--a', '99'], strict=False)

        assert reg.select()['a'] == 99  # COMMANDLINE_ONLY ignores config


class TestListTypes:
    """Tests for list type entries"""

    def test_list_from_config(self, temp_config_dir):
        """Test: list values from config file"""
        from config_reg.type_def import ConfigEntryCommandlineSeqPattern

        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'items': [1, 2, 3]})

        reg = ConfigRegistry()
        reg.register('items',
                     category=list[int],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineSeqPattern.COMMA_SEP)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path], strict=False)

        assert reg.select()['items'] == [1, 2, 3]

    def test_list_cmdline_comma_sep(self, temp_config_dir):
        """Test: list from cmdline with comma separator"""
        from config_reg.type_def import ConfigEntryCommandlineSeqPattern

        reg = ConfigRegistry()
        reg.register('items',
                     category=list[int],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineSeqPattern.COMMA_SEP)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['--items', '1,2,3,4'], strict=False)

        assert reg.select()['items'] == [1, 2, 3, 4]

    def test_list_cmdline_colon_sep(self, temp_config_dir):
        """Test: list from cmdline with colon separator"""
        from config_reg.type_def import ConfigEntryCommandlineSeqPattern

        reg = ConfigRegistry()
        reg.register('paths',
                     category=list[str],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineSeqPattern.COLON_SEP)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['--paths', '/path/a:/path/b:/path/c'], strict=False)

        assert reg.select()['paths'] == ['/path/a', '/path/b', '/path/c']

    def test_list_cmdline_override_config(self, temp_config_dir):
        """Test: cmdline list overrides config list"""
        from config_reg.type_def import ConfigEntryCommandlineSeqPattern

        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'nums': [10, 20, 30]})

        reg = ConfigRegistry()
        reg.register('nums',
                     category=list[int],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineSeqPattern.COMMA_SEP)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '--nums', '1,2'], strict=False)

        assert reg.select()['nums'] == [1, 2]  # cmdline overrides config

    def test_list_config_override_cmdline(self, temp_config_dir):
        """Test: config list overrides earlier cmdline list"""
        from config_reg.type_def import ConfigEntryCommandlineSeqPattern

        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'nums': [10, 20, 30]})

        reg = ConfigRegistry()
        reg.register('nums',
                     category=list[int],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineSeqPattern.COMMA_SEP)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['--nums', '1,2', '-c', cfg1_path], strict=False)

        assert reg.select()['nums'] == [10, 20, 30]  # config overrides cmdline

    def test_list_string_type(self, temp_config_dir):
        """Test: list of strings from config"""
        from config_reg.type_def import ConfigEntryCommandlineSeqPattern

        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'tags': ['alpha', 'beta', 'gamma']})

        reg = ConfigRegistry()
        reg.register('tags',
                     category=list[str],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineSeqPattern.COMMA_SEP)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path], strict=False)

        assert reg.select()['tags'] == ['alpha', 'beta', 'gamma']

    def test_list_empty_from_config(self, temp_config_dir):
        """Test: empty list from config"""
        from config_reg.type_def import ConfigEntryCommandlineSeqPattern

        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'items': []})

        reg = ConfigRegistry()
        reg.register('items',
                     category=list[int],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineSeqPattern.COMMA_SEP)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path], strict=False)

        assert reg.select()['items'] == []

    def test_nested_list(self, temp_config_dir):
        """Test: list in nested key"""
        from config_reg.type_def import ConfigEntryCommandlineSeqPattern

        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'data': {'ids': [1, 2, 3]}})

        reg = ConfigRegistry()
        reg.register('data.ids',
                     category=list[int],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineSeqPattern.COMMA_SEP)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '--data.ids', '4,5,6'], strict=False)

        assert reg.select()['data']['ids'] == [4, 5, 6]


class TestDictTypes:
    """Tests for dict type entries"""

    def test_dict_from_config(self, temp_config_dir):
        """Test: dict values from config file"""
        from config_reg.type_def import ConfigEntryCommandlineMapPattern

        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'params': {'a': 1, 'b': 2}})

        reg = ConfigRegistry()
        reg.register('params',
                     category=dict[str, int],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineMapPattern.COMMA_SEP_COLON)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path], strict=False)

        assert reg.select()['params'] == {'a': 1, 'b': 2}

    def test_dict_cmdline_comma_colon(self, temp_config_dir):
        """Test: dict from cmdline with comma-colon pattern"""
        from config_reg.type_def import ConfigEntryCommandlineMapPattern

        reg = ConfigRegistry()
        reg.register('params',
                     category=dict[str, int],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineMapPattern.COMMA_SEP_COLON)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['--params', 'x:10,y:20,z:30'], strict=False)

        assert reg.select()['params'] == {'x': 10, 'y': 20, 'z': 30}

    def test_dict_cmdline_override_config(self, temp_config_dir):
        """Test: cmdline dict overrides config dict"""
        from config_reg.type_def import ConfigEntryCommandlineMapPattern

        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'opts': {'key1': 'val1', 'key2': 'val2'}})

        reg = ConfigRegistry()
        reg.register('opts',
                     category=dict[str, str],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineMapPattern.COMMA_SEP_COLON)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '--opts', 'new:value'], strict=False)

        assert reg.select()['opts'] == {'new': 'value'}  # cmdline overrides config

    def test_dict_config_override_cmdline(self, temp_config_dir):
        """Test: config dict overrides earlier cmdline dict"""
        from config_reg.type_def import ConfigEntryCommandlineMapPattern

        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'opts': {'from': 'config'}})

        reg = ConfigRegistry()
        reg.register('opts',
                     category=dict[str, str],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineMapPattern.COMMA_SEP_COLON)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['--opts', 'from:cmdline', '-c', cfg1_path], strict=False)

        assert reg.select()['opts'] == {'from': 'config'}  # config overrides cmdline

    def test_dict_string_value(self, temp_config_dir):
        """Test: dict with string values from config"""
        from config_reg.type_def import ConfigEntryCommandlineMapPattern

        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'env': {'HOME': '/home/user', 'PATH': '/usr/bin'}})

        reg = ConfigRegistry()
        reg.register('env',
                     category=dict[str, str],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineMapPattern.COMMA_SEP_COLON)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path], strict=False)

        assert reg.select()['env'] == {'HOME': '/home/user', 'PATH': '/usr/bin'}

    def test_dict_empty_from_config(self, temp_config_dir):
        """Test: empty dict from config"""
        from config_reg.type_def import ConfigEntryCommandlineMapPattern

        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'params': {}})

        reg = ConfigRegistry()
        reg.register('params',
                     category=dict[str, int],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineMapPattern.COMMA_SEP_COLON)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path], strict=False)

        assert reg.select()['params'] == {}

    def test_nested_dict(self, temp_config_dir):
        """Test: dict in nested key"""
        from config_reg.type_def import ConfigEntryCommandlineMapPattern

        cfg1_path = create_yaml_file(temp_config_dir, 'cfg1.yaml', {'config': {'options': {'a': 1, 'b': 2}}})

        reg = ConfigRegistry()
        reg.register('config.options',
                     category=dict[str, int],
                     source=ConfigEntrySource.COMMANDLINE_OVER_CONFIG,
                     cmdpattern=ConfigEntryCommandlineMapPattern.COMMA_SEP_COLON)

        parser = argparse.ArgumentParser()
        reg.hook(parser)

        reg.parse(parser, ['-c', cfg1_path, '--config.options', 'c:3,d:4'], strict=False)

        assert reg.select()['config']['options'] == {'c': 3, 'd': 4}



"""
Unit tests for split_argv_by_cfg function
"""
import pytest
from config_reg.reg import split_argv_by_cfg


class TestSplitArgvByCfg:
    """Tests for split_argv_by_cfg function"""

    def test_basic_with_short_flag(self):
        """Test basic case with -c flag"""
        result = split_argv_by_cfg(['--a', '1', '-c', 'f.yaml', '--b', '2'])
        expected = [('args', ['--a', '1']), ('cfg', 'f.yaml'), ('args', ['--b', '2'])]
        assert result == expected

    def test_basic_with_long_flag(self):
        """Test basic case with --cfg flag"""
        result = split_argv_by_cfg(['--a', '1', '--cfg', 'f.yaml', '--b', '2'])
        expected = [('args', ['--a', '1']), ('cfg', 'f.yaml'), ('args', ['--b', '2'])]
        assert result == expected

    def test_cfg_equals_format(self):
        """Test --cfg= format"""
        result = split_argv_by_cfg(['--a', '1', '--cfg=f.yaml', '--b', '2'])
        expected = [('args', ['--a', '1']), ('cfg', 'f.yaml'), ('args', ['--b', '2'])]
        assert result == expected

    def test_cfg_at_start(self):
        """Test config file at the beginning"""
        result = split_argv_by_cfg(['-c', 'f.yaml', '--a', '1'])
        expected = [('cfg', 'f.yaml'), ('args', ['--a', '1'])]
        assert result == expected

    def test_cfg_at_end(self):
        """Test config file at the end"""
        result = split_argv_by_cfg(['--a', '1', '-c', 'f.yaml'])
        expected = [('args', ['--a', '1']), ('cfg', 'f.yaml')]
        assert result == expected

    def test_only_config_files(self):
        """Test with only config files"""
        result = split_argv_by_cfg(['-c', 'f1.yaml', '--cfg', 'f2.yaml'])
        expected = [('cfg', 'f1.yaml'), ('cfg', 'f2.yaml')]
        assert result == expected

    def test_only_args(self):
        """Test with only arguments"""
        result = split_argv_by_cfg(['--a', '1', '--b', '2'])
        expected = [('args', ['--a', '1', '--b', '2'])]
        assert result == expected

    def test_empty_list(self):
        """Test with empty list"""
        result = split_argv_by_cfg([])
        expected = []
        assert result == expected

    def test_multiple_configs_interleaved(self):
        """Test multiple config files interleaved with arguments"""
        result = split_argv_by_cfg(['--a', '1', '-c', 'f1.yaml', '--b', '2', '--cfg=f2.yaml', '--c', '3'])
        expected = [('args', ['--a', '1']), ('cfg', 'f1.yaml'), ('args', ['--b', '2']), ('cfg', 'f2.yaml'),
                    ('args', ['--c', '3'])]
        assert result == expected

    def test_consecutive_configs(self):
        """Test consecutive config files without args in between"""
        result = split_argv_by_cfg(['--a', '1', '-c', 'f1.yaml', '-c', 'f2.yaml', '--b', '2'])
        expected = [('args', ['--a', '1']), ('cfg', 'f1.yaml'), ('cfg', 'f2.yaml'), ('args', ['--b', '2'])]
        assert result == expected

    def test_missing_config_path_short(self):
        """Test missing config file path after -c"""
        with pytest.raises(ValueError, match="Missing config file path after -c"):
            split_argv_by_cfg(['--a', '1', '-c'])

    def test_missing_config_path_long(self):
        """Test missing config file path after --cfg"""
        with pytest.raises(ValueError, match="Missing config file path after --cfg"):
            split_argv_by_cfg(['--a', '1', '--cfg'])

    def test_empty_cfg_equals(self):
        """Test empty path in --cfg="""
        with pytest.raises(ValueError, match="Empty config file path"):
            split_argv_by_cfg(['--a', '1', '--cfg='])

    def test_mixed_formats(self):
        """Test mixing -c, --cfg, and --cfg= formats"""
        result = split_argv_by_cfg(['-c', 'a.yaml', '--cfg', 'b.yaml', '--cfg=c.yaml'])
        expected = [('cfg', 'a.yaml'), ('cfg', 'b.yaml'), ('cfg', 'c.yaml')]
        assert result == expected

    def test_config_with_path(self):
        """Test config file with path"""
        result = split_argv_by_cfg(['-c', '/path/to/config.yaml', '--a', '1'])
        expected = [('cfg', '/path/to/config.yaml'), ('args', ['--a', '1'])]
        assert result == expected

    def test_config_with_relative_path(self):
        """Test config file with relative path"""
        result = split_argv_by_cfg(['-c', '../configs/base.yml', '--b', '2'])
        expected = [('cfg', '../configs/base.yml'), ('args', ['--b', '2'])]
        assert result == expected

    def test_double_dash_terminator(self):
        """Test -- stops config file parsing"""
        result = split_argv_by_cfg(['--a', '1', '--', '-c', 'not_config', '--cfg', 'also_not'])
        expected = [('args', ['--a', '1', '--', '-c', 'not_config', '--cfg', 'also_not'])]
        assert result == expected

    def test_double_dash_with_cfg_before(self):
        """Test -- after config file"""
        result = split_argv_by_cfg(['-c', 'config.yaml', '--', '-c', 'not_config'])
        expected = [('cfg', 'config.yaml'), ('args', ['--', '-c', 'not_config'])]
        assert result == expected

    def test_double_dash_at_start(self):
        """Test -- at the beginning"""
        result = split_argv_by_cfg(['--', '-c', 'not_config', '--a', '1'])
        expected = [('args', ['--', '-c', 'not_config', '--a', '1'])]
        assert result == expected

    def test_suspicious_option_warning(self, caplog):
        """Test warning when config file path looks like an option"""
        import logging
        with caplog.at_level(logging.WARNING):
            result = split_argv_by_cfg(['-c', '--verbose'])
        assert result == [('cfg', '--verbose')]
        assert "looks like an option" in caplog.text

    def test_suspicious_option_short_flag_warning(self, caplog):
        """Test warning for short option as config path"""
        import logging
        with caplog.at_level(logging.WARNING):
            result = split_argv_by_cfg(['--cfg', '-v'])
        assert result == [('cfg', '-v')]
        assert "looks like an option" in caplog.text

    def test_cfg_equals_suspicious_warning(self, caplog):
        """Test warning for --cfg= with option-like value"""
        import logging
        with caplog.at_level(logging.WARNING):
            result = split_argv_by_cfg(['--cfg=--debug'])
        assert result == [('cfg', '--debug')]
        assert "looks like an option" in caplog.text

    def test_cfg_after_double_dash_warning(self, caplog):
        """Test warning when -c appears after --"""
        import logging
        with caplog.at_level(logging.WARNING):
            result = split_argv_by_cfg(['--', '-c', 'file.yaml'])
        assert result == [('args', ['--', '-c', 'file.yaml'])]
        assert "after '--'" in caplog.text



from __future__ import annotations

import os
import typing

from .type_def import ConfigEntryValueUnspecified


class ConfigEntryCallback:
    dependency: list[str] = []
    always: bool = False

    def __call__(self, curr_key: str, curr_value: typing.Any, prog: str, dep: typing.Mapping) -> typing.Any:
        return ConfigEntryValueUnspecified


def resolve_callback_dependency(callback_map: typing.Mapping[str, ConfigEntryCallback]) -> typing.Optional[list[str]]:
    """Resolve callback dependencies using topological sort.
    
    Args:
        callback_map: Mapping of key names to their callback objects.
        
    Returns:
        A list of keys in topologically sorted order (dependencies first),
        or None if there is a cyclic dependency.
    """
    # obtain adj list
    all_keys = set(callback_map.keys())
    adj_list: dict[str, list[str]] = {}
    for k, v in callback_map.items():
        dlist = []
        for d in v.dependency:
            if d in all_keys:
                dlist.append(d)
        adj_list[k] = dlist

    # get proc requirement
    proc_dep_list = adj_list.copy()
    proc_clear_list: dict[str, list[str]] = {}
    for k, dlist in proc_dep_list.items():
        for vv in dlist:
            if vv not in proc_clear_list:
                proc_clear_list[vv] = []
            proc_clear_list[vv].append(k)

    # toposort
    res = []
    while True:
        if len(proc_dep_list) == 0:
            break

        # iterate over proc_dep_list to find whether have a 0
        k_curr = next(iter(proc_dep_list.keys()))
        indeg_curr = len(proc_dep_list[k_curr])
        for _k in proc_dep_list:
            _indeg = len(proc_dep_list[_k])
            if _indeg < indeg_curr:
                k_curr = _k
                indeg_curr = _indeg
                if indeg_curr == 0:
                    break

        if indeg_curr > 0:
            return None

        res.append(k_curr)

        if k_curr in proc_clear_list:
            clear_list = proc_clear_list[k_curr]
            for k_node in clear_list:
                proc_dep_list[k_node].remove(k_curr)
            proc_clear_list.pop(k_curr)

        proc_dep_list.pop(k_curr)
    return res


# common callbacks
class AbspathCallback(ConfigEntryCallback):
    always: bool = True

    def __init__(self) -> None:
        super().__init__()
        self.curr_path = os.getcwd()

    def to_abs(self, v):
        if os.path.isabs(v):
            return os.path.normcase(os.path.normpath(v))
        else:
            return os.path.normcase(os.path.normpath(os.path.join(self.curr_path, v)))

    def __call__(self, curr_key: str, curr_value: typing.Any, prog: str, dep: typing.Mapping) -> typing.Any:
        if curr_value == ConfigEntryValueUnspecified or curr_value is None:
            return curr_value

        if isinstance(curr_value, list):
            return [self.to_abs(e) for e in curr_value]
        else:
            return self.to_abs(curr_value)


abspath_callback = AbspathCallback()


class InterpolationCallback(ConfigEntryCallback):
    always: bool = False

    def __init__(self, template: str, ensure_str=True) -> None:
        super().__init__()

        from .util import subst_util

        self.template = template
        self.ensure_str = ensure_str

        # extract dependency
        key_list, span_list = subst_util.extract_special_part(self.template)
        self.key_list = key_list
        self.span_list = span_list
        self.dependency = self.key_list.copy()

    def __call__(self, curr_key: str, curr_value: typing.Any, prog: str, dep: typing.Mapping) -> typing.Any:
        from .util import subst_util
        from .index import index_key
        if curr_key in self.key_list:
            raise ValueError(f"Cyclic dependency detected: key '{curr_key}' references itself in template")

        replacement_list = []
        for k in self.key_list:
            if k not in dep:
                raise KeyError(f"Key '{k}' not found in dependency, got dep keys: {list(dep.keys())}")
            v = dep[k]
            if self.ensure_str:
                if not isinstance(v, str):
                    raise TypeError(f"Expected str value for key '{k}', got {type(v).__name__}: {v}")
            replacement_list.append(str(v))

        new_value = subst_util.replace_from_span(self.template, self.span_list, replacement_list)
        return new_value



from .type_def import ConfigEntrySource, ConfigEntryValueUnspecified
from .type_def import (
    ConfigEntryCommandlinePattern,
    ConfigEntryCommandlineSeqPattern,
    ConfigEntryCommandlineMapPattern,
    ConfigEntryCommandlineBoolPattern,
)
from .callback import ConfigEntryCallback
from .reg import ConfigRegistry

from .convert import instantiate

__all__ = [
    "ConfigRegistry",
    "ConfigEntrySource",
    "ConfigEntryCommandlinePattern",
    "ConfigEntryCommandlineSeqPattern",
    "ConfigEntryCommandlineMapPattern",
    "ConfigEntryCommandlineBoolPattern",
    "ConfigEntryValueUnspecified",
    "ConfigEntryCallback",
    "instantiate",
]



import copy
import functools
from dataclasses import is_dataclass, fields
from enum import Enum
from textwrap import dedent
from typing import Any, Callable, Dict, List, Sequence, Tuple, Union


class InstantiationException(Exception):
    pass


class _Keys(str, Enum):
    TARGET = "_target_"
    RECURSIVE = "_recursive_"
    ARGS = "_args_"
    PARTIAL = "_partial_"


def _is_target(x: Any) -> bool:
    return isinstance(x, dict) and "_target_" in x


def _extract_pos_args(input_args: Any, kwargs: Any) -> Tuple[Any, Any]:
    config_args = kwargs.pop(_Keys.ARGS, ())
    output_args = config_args
    if isinstance(config_args, Sequence):
        if len(input_args) > 0:
            output_args = input_args
    else:
        raise InstantiationException(f"Unsupported _args_ type: '{type(config_args).__name__}'. value: '{config_args}'")
    return output_args, kwargs


def _call_target(_target_: Callable[..., Any], _partial_: bool, args: Tuple[Any, ...], kwargs: Dict[str, Any],
                 full_key: str) -> Any:
    try:
        args, kwargs = _extract_pos_args(args, kwargs)
    except Exception as e:
        msg = f"Error in collecting args and kwargs for '{_convert_target_to_string(_target_)}': {repr(e)}"
        if full_key:
            msg += f"\nfull_key: {full_key}"
        raise InstantiationException(msg) from e

    if _partial_:
        try:
            return functools.partial(_target_, *args, **kwargs)
        except Exception as e:
            msg = f"Error in creating partial({_convert_target_to_string(_target_)}, ...) object: {repr(e)}"
            if full_key:
                msg += f"\nfull_key: {full_key}"
            raise InstantiationException(msg) from e
    else:
        try:
            return _target_(*args, **kwargs)
        except Exception as e:
            msg = f"Error in call to target '{_convert_target_to_string(_target_)}': {repr(e)}"
            if full_key:
                msg += f"\nfull_key: {full_key}"
            raise InstantiationException(msg) from e


def _convert_target_to_string(t: Any) -> Any:
    if callable(t):
        return f"{t.__module__}.{t.__qualname__}"
    else:
        return str(t)


def _is_dataclass_instance(obj: Any) -> bool:
    """Check if obj is a dataclass instance (not a dataclass class)."""
    return is_dataclass(obj) and not isinstance(obj, type)


def _dataclass_to_dict_shallow(obj: Any) -> Dict[str, Any]:
    """Shallow convert a dataclass instance to dict. Child values are kept as-is."""
    return {f.name: getattr(obj, f.name) for f in fields(obj)}


def _resolve_target(target: Union[str, type, Callable[..., Any]], full_key: str) -> Union[type, Callable[..., Any]]:
    """Resolve target string, type or callable into type or callable."""
    if isinstance(target, str):
        try:
            target = _locate(target)
        except Exception as e:
            msg = f"Error locating target '{target}'"
            if full_key:
                msg += f"\nfull_key: {full_key}"
            raise InstantiationException(msg) from e
    if not callable(target):
        msg = f"Expected a callable target, got '{target}' of type '{type(target).__name__}'"
        if full_key:
            msg += f"\nfull_key: {full_key}"
        raise InstantiationException(msg)
    return target


def _locate(path: str) -> Any:
    """
    Locate an object by name or dotted path, importing as necessary.
    This is similar to the pydoc function `locate`, except that it checks for
    the module from the given path from back to front.
    """
    if path == "":
        raise ImportError("Empty path")
    from importlib import import_module
    from types import ModuleType

    parts = [part for part in path.split(".")]
    for part in parts:
        if not len(part):
            raise ValueError(f"Error loading '{path}': invalid dotstring." + "\nRelative imports are not supported.")
    if len(parts) == 0:
        raise ValueError(f"Error loading '{path}': path resulted in no parts after splitting.")
    part0 = parts[0]
    try:
        obj = import_module(part0)
    except Exception as exc_import:
        raise ImportError(f"Error loading '{path}':\n{repr(exc_import)}" +
                          f"\nAre you sure that module '{part0}' is installed?") from exc_import
    for m in range(1, len(parts)):
        part = parts[m]
        try:
            obj = getattr(obj, part)
        except AttributeError as exc_attr:
            parent_dotpath = ".".join(parts[:m])
            if isinstance(obj, ModuleType):
                mod = ".".join(parts[:m + 1])
                try:
                    obj = import_module(mod)
                    continue
                except ModuleNotFoundError as exc_import:
                    raise ImportError(
                        f"Error loading '{path}':\n{repr(exc_import)}" +
                        f"\nAre you sure that '{part}' is importable from module '{parent_dotpath}'?") from exc_import
                except Exception as exc_import:
                    raise ImportError(f"Error loading '{path}':\n{repr(exc_import)}") from exc_import
            raise ImportError(f"Error loading '{path}':\n{repr(exc_attr)}" +
                              f"\nAre you sure that '{part}' is an attribute of '{parent_dotpath}'?") from exc_attr
    return obj


def instantiate_node(config: Any, *args: Any, recursive: bool = True, partial: bool = False, full_key: str = "") -> Any:
    if config is None:
        return None

    # Shallow convert dataclass to dict (lazy: children remain as-is)
    if _is_dataclass_instance(config):
        config = _dataclass_to_dict_shallow(config)

    if not isinstance(config, (list, dict)):
        return config

    if not isinstance(recursive, bool):
        msg = f"Instantiation: _recursive_ flag must be a bool, got {type(recursive)}"
        if full_key:
            msg += f"\nfull_key: {full_key}"
        raise TypeError(msg)

    if not isinstance(partial, bool):
        msg = f"Instantiation: _partial_ flag must be a bool, got {type(partial)}"
        if full_key:
            msg += f"\nfull_key: {full_key}"
        raise TypeError(msg)

    # main handling
    if isinstance(config, list):
        items = [instantiate_node(item, recursive=recursive) for item in config]
        return items

    elif isinstance(config, dict):
        exclude_keys = set({_Keys.TARGET, _Keys.RECURSIVE, _Keys.PARTIAL})
        if _is_target(config):
            _target_ = _resolve_target(config.get(_Keys.TARGET, ""), full_key)

            # Node-level _recursive_ and _partial_ override
            node_recursive = config.get(_Keys.RECURSIVE, recursive)
            node_partial = config.get(_Keys.PARTIAL, partial)

            kwargs = {}
            for key in config.keys():
                if key not in exclude_keys:
                    value = config[key]
                    if node_recursive:
                        value = instantiate_node(value, recursive=node_recursive)
                    else:
                        # Even with recursive=False, convert dataclass to dict (shallow)
                        if _is_dataclass_instance(value):
                            value = _dataclass_to_dict_shallow(value)
                    kwargs[key] = value
            return _call_target(_target_, node_partial, args, kwargs, full_key)

        else:
            return {k: instantiate_node(v, recursive=recursive) for k, v in config.items()}

    else:
        raise InstantiationException(f"Unexpected config type: {type(config).__name__}")


def _get_target_from_config(config: Any) -> Any:
    """Get _target_ from config (dict, dataclass, or via attribute)."""
    if isinstance(config, dict):
        return config.get("_target_")
    elif _is_dataclass_instance(config):
        return getattr(config, "_target_", None)
    elif hasattr(config, "_target_"):
        return getattr(config, "_target_", None)
    return None


def instantiate(config: Any, *args: Any, full_key: str = "", **kwargs: Any) -> Any:
    if config is None:
        return None

    # Check _target_ exists (support dict and dataclass)
    target = _get_target_from_config(config)
    if target is None or target == "":
        raise InstantiationException(
            dedent(f"""\
            Config has missing value for key `_target_`, cannot instantiate.
            Config type: {type(config).__name__}
            Check that the `_target_` key in your dataclass is properly annotated and overridden.
        """))

    # Shallow convert dataclass to dict at top level
    if _is_dataclass_instance(config):
        config = _dataclass_to_dict_shallow(config)

    if isinstance(config, dict):
        config_copy = copy.deepcopy(config)
        if kwargs:
            config_copy.update(kwargs)

        _recursive_ = config_copy.pop(_Keys.RECURSIVE, True)
        _partial_ = config_copy.pop(_Keys.PARTIAL, False)

        return instantiate_node(
            config_copy,
            *args,
            recursive=_recursive_,
            partial=_partial_,
            full_key=full_key,
        )

    elif isinstance(config, list):
        config_copy = copy.deepcopy(config)
        return instantiate_node(
            config_copy,
            *args,
            recursive=kwargs.pop(_Keys.RECURSIVE, True),
            partial=kwargs.pop(_Keys.PARTIAL, False),
            full_key=full_key,
        )

    else:
        raise InstantiationException(
            dedent(f"""\
            Cannot instantiate config of type {type(config).__name__}.
            Top level config must be a plain dict/list, or a structured config class or instance."""))
